from .text_classification import TextClassificationDataModule
from .text_pair_classification import TextPairClassificationDataModule
from .text_pair_regression import TextPairRegressionDataModule
from .token_classification import TokenClassificationDataModule
from .span_extraction import SpanExtractionDataModule
from .relation_extraction import RelationExtractionDataModule
from .prompt_span_extraction import PromptSpanExtractionDataModule
from .prompt_relation_extraction import PromptRelationExtractionDataModule
from .event_extraction import EventExtractionDataModule
from .entity_extraction import EntityExtractionDataModule
from .question_answering import QuestionAnsweringDataModule
